/**********************************************************************
 
	Copyright (C) 2007 Hirohisa MORI <joshua@globalbase.org>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/


#ifndef ___AUTHENTIFICATION_H___
#define ___AUTHENTIFICATION_H___

#include	"machine/include.h"
#include	"long_char.h"
#include	"xl.h"

typedef struct auth_key_info {
	int			type;
#define KT_RANDOM		1
#define KT_PUBLIC_KEY		2
	int			status;
#define KS_NOTUSE		0
#define KS_USE			1
#define KS_WAIT_TIMEOUT		2
#define KS_DONTUSE		3
	int			key_id;
	unsigned char *		public_key;
	unsigned char *		private_key;
	INTEGER64		create_time;
	INTEGER64		expire_time;
} AUTH_KEY_INFO;

typedef struct auth_status {
	short			app_allow;
	short			app_deny;
	short			gateway_allow;
	short			gateway_denay;
#define AF_VIEW			0x0001
#define AF_COPY			0x0002
#define AF_EXPORT		0x0004
#define AF_CACHE		0x0008
#define AF_VERSION		0x0010
#define AF_USER1		0x0020
#define AF_USER2		0x0040
#define AF_ALL			(AF_VIEW|AF_COPY|AF_EXPORT|AF_CACHE|AF_VERSION|AF_USER1|AF_USER2)
} AUTH_STATUS;

typedef struct auth_user {
	L_CHAR *		name;
	int			type;
/* AT_TYPE */
	int			uid;
	int			key_id;
	AUTH_KEY_INFO *		key_info;
	L_CHAR *		passwd;
	INTEGER64		expire_time;
	AUTH_STATUS		sts;
} AUTH_USER;

typedef struct auth_type_header {
	int			id;
	struct auth_table *	tbl;
	int			type;
#define AT_OPT_USER_PASSWD	1
#define AT_OPT_ADDRESS		2
#define AT_OPT_KEY		3
#define AT_OPT_VERSION		4
#define AT_OPT_LIST		5

#define AT_D(x,y,z)	((x)|((y)<<8)|(z)<<16)

#define AT_CLOSE		AT_D(0,0,0)
#define AT_TMP			AT_D(1,0,0)
#define AT_FIX_USER_PASSWD	AT_D(2,AT_OPT_USER_PASSWD,AT_OPT_USER_PASSWD)
#define AT_PAM_USER_PASSWD	AT_D(3,0,AT_OPT_USER_PASSWD)
#define AT_ADDRESS		AT_D(4,AT_OPT_ADDRESS,AT_OPT_ADDRESS)
#define AT_PUBLIC_KEY		AT_D(5,0,AT_OPT_KEY)
#define AT_CLOSED_KEY		AT_D(6,0,AT_OPT_KEY)
#define AT_PREPAID		AT_D(7,0,AT_OPT_KEY)
#define AT_SEQURE_PATH		AT_D(8,0,0)
#define AT_VERSION		AT_D(9,AT_OPT_VERSION,AT_OPT_VERSION)

#define AT_AND			AT_D(0x80,AT_OPT_LIST,0)
#define AT_OR			AT_D(0x81,AT_OPT_LIST,0)
	AUTH_STATUS		sts;
} AUTH_TYPE_HEADER;


typedef struct auth_type_list {
	AUTH_TYPE_HEADER	h;
	int			len;
	union auth_type **	list;
} AUTH_TYPE_LIST;

typedef struct auth_type_user_passwd {
	AUTH_TYPE_HEADER	h;
	L_CHAR *		user;
	L_CHAR *		passwd;
} AUTH_TYPE_USER_PASSWD;

typedef struct auth_type_address {
	AUTH_TYPE_HEADER	h;
	ACCESS_PERMISSION	ap;
} AUTH_TYPE_ADDRESS;

typedef struct auth_type_version {
	AUTH_TYPE_HEADER	h;
	L_CHAR *		version;
} AUTH_TYPE_VERSION;

typedef union auth_type {
	AUTH_TYPE_HEADER	h;
	AUTH_TYPE_USER_PASSWD	pw;
	AUTH_TYPE_ADDRESS	ad;
	AUTH_TYPE_LIST		lst;
} AUTH_TYPE;

typedef struct auth_prefix {
	struct auth_prefix *	next;
	L_CHAR *		prefix;
	AUTH_TYPE *		at_type;
} AUTH_PREFIX;

typedef struct auth_dir {
	L_CHAR *		dir_name;
	L_CHAR *		total_path;
	struct auth_dir *	child_list;
	struct auth_dir *	child_next;
	AUTH_PREFIX *		prefix;
} AUTH_DIR;

typedef struct auth_result_header {
	union auth_result *	next;
	AUTH_TYPE *		type;
	AUTH_USER *		user_info;
	int			status;
#define ATS_NON_AUTH		1
#define ATS_OK			2
#define ATS_REJECT		3
	AUTH_STATUS		sts;
} AUTH_RESULT_HEADER;


typedef struct auth_result_user_passwd {
	AUTH_RESULT_HEADER	h;
	L_CHAR *		user;
	L_CHAR *		passwd;
} AUTH_RESULT_USER_PASSWD;

typedef struct auth_result_address {
	AUTH_RESULT_HEADER	h;
	HOST_ADDR		addr;
} AUTH_RESULT_ADDRESS;

typedef struct auth_result_key {
	AUTH_RESULT_HEADER	h;
	unsigned char *		public_key;
	unsigned char *		private_key;
	AUTH_KEY_INFO *		key_info;
	INTEGER64		remain;
} AUTH_RESULT_KEY;


typedef struct auth_result_version {
	AUTH_RESULT_HEADER	h;
	L_CHAR *		version;
} AUTH_RESULT_VERSION;


typedef union auth_result {
	AUTH_RESULT_HEADER	h;
	AUTH_RESULT_USER_PASSWD	pw;
	AUTH_RESULT_ADDRESS	ad;
	AUTH_RESULT_KEY		key;
} AUTH_RESULT;

typedef struct auth_set {
	AUTH_KEY_INFO *		auth_key_list;
	AUTH_USER *		auth_user_list;
	AUTH_TYPE *		auth_type_list;
	AUTH_DIR *		auth_dir_tree;
	AUTH_RESULT *		client_result;
	unsigned		enable_client_result:1;
} AUTH_SET;

typedef struct auth_connection_set {
	AUTH_SET *		base_set;
	AUTH_RESULT *		result;
	XL_INTERPRETER *	xli;
} AUTH_CONNECTION_SET;

typedef struct auth_table {
	int			id;
	char *			name;
	AUTH_TYPE *		(*new_type)(AUTH_SET*,void *);
	XL_SEXP *		(*to_type_sexp)(AUTH_TYPE*);
	AUTH_TYPE *		(*to_type)(AUTH_SET *,XL_SEXP*);
	void			(*free_type)(AUTH_TYPE*,int);
#define AT_FREE_TYPE_ALL	1
#define AT_FREE_TYPE_DUMMY	2
	void			(*ctl_type)(AUTH_TYPE*,int,void*);
	
	AUTH_RESULT *		(*new_result)(void*);
	XL_SEXP *		(*to_result_sexp)(AUTH_RESULT*);
	AUTH_RESULT *		(*to_result)(AUTH_TYPE *,XL_SEXP*);
	void			(*free_result)(AUTH_RESULT*);
	void			(*clt_result)(AUTH_TYPE*,int,void*);

#define AT_CTL_INSERT		1
#define AT_CTL_DELETE		2

	AUTH_RESULT *		(*authentification)(AUTH_TYPE*,AUTH_CONNECTION_SET *,int level,AUTH_RESULT *);
//	level = 0	server can autholize only connection information
//	level = 1	matching authentification at client
//	level = 2	server can autholize using return values from client
//	level = 3	server and client must has transaction

} AUTH_TABLE;



typedef struct at_list_t {
	AUTH_STATUS		sts;
	AUTH_TYPE *		d;
} AT_LIST_T;


#endif



