/**********************************************************************
 
	Copyright (C) 2007-
	Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/

#include	"randomkey.h"
#include	"memory_debug.h"

unsigned char
gen_random_digit()
{
	return (rand())%RND_ASC_DIGIT;
}

void
gen_continue_random_digit(unsigned char * str,int size)
{
int i;
	for ( i = 0 ; i < size ; i++ )
		*str++ = gen_random_digit();
	*str = 0;
}

unsigned char
r_add(unsigned char a,unsigned char b)
{
	return (a+b)%RND_ASC_DIGIT;
}

unsigned char
r_mul(unsigned char a,unsigned char b)
{
	return (a*b)%RND_ASC_DIGIT;
}

void
gen_hafman_rnd_matrix(unsigned char * mtx,int size,int distance)
{
int width,height;
unsigned char ** row,**other_row;;
int i,j,k;
unsigned char p;
	width = HF_RND_MTX_X(size,distance);
	height = HF_RND_MTX_Y(size,distance);
	memset(mtx,0,width*height);
	row = d_alloc(height*sizeof(char*));
	other_row = &row[width];
	for ( i = 0 ; i < distance-1 ; i ++ ) {
		for ( ; ; ) {
			p = rand()%height;
			if ( mtx[p*width] == 1 )
				continue;
			break;
		}
		mtx[p*width] = 1;
	}
	j = 0;
	k = 0;
	i = 0;
	for ( ; i < height ; i ++ ) {
		if ( mtx[i*width] ) {
			other_row[k++] = &mtx[i*width];
		}
		else {
			row[j++] = &mtx[i*width];
		}
	}
	for ( i = 0 ; i < width ; i ++ ) {
		for ( j = 0 ; j < i ; j ++ )
			row[i][j] = 0;
		p = 0;
		for ( ; p ; )
			p = gen_random_digit();
		row[i][i] = p;
		for ( j = i+1 ; j < width ; j ++ )
			row[i][j] = gen_random_digit();
	}
	for ( i = width-1 ; i >= 0 ; i -- ) {
		for ( k = i+1 ; k < width ; k ++ ) {
			p = gen_random_digit();
			for ( j = 0 ; j < width ; j ++ ) {
				row[k][j] = 
					r_add(r_mul(p,row[i][j]),row[k][j]);
			}
		}
	}
	
	for ( i = 0 ; i < height-width ; i ++ ) {
		for ( k = 0 ; k < width ; k ++ ) {
			p = gen_random_digit();
			for ( j = 0 ; j < width ; j ++ ) {
				other_row[i][j] = r_add(r_mul(p,row[k][j]),
							other_row[i][j]);
			}
		}
	}
	d_f_ree(row);
}


void
get_hafman_random_code(unsigned char *result,unsigned char * mtx,unsigned char * str,int width,int height)
{
int i,j;
unsigned char p;
	for ( i = 0 ; i < height ; i ++ ) {
		p = 0;
		for ( j = 0 ; j < width ; j ++ )
			p = r_add(p,r_mul(mtx[j+i*width],str[j]));
		result[i] = p;
	}
}

unsigned char
r_asc_bin_to_str_1(unsigned char a)
{
	if ( a < 9 )
		return a+'1';
	a -= 9;
	if ( a < 26 )
		return a+'a';
	a -= 26;
	if ( a < 26 )
		return a+'A';
	return '?';
}

unsigned char
r_asc_str_to_bin_1(unsigned char a)
{
	if ( '1' <= a && a <= '9' )
		return a - '1';
	if ( 'a' <= a && a <= 'z' )
		return a - 'a' + 9;
	if ( 'A' <= a && a <= 'Z' )
		return a - 'A' + 9 + 26;
	return 0xff;
}

void
r_asc_bin_to_str(unsigned char * result,unsigned char * str)
{
	for ( ; *str < RND_ASC_DIGIT ; )
		*result++ = r_asc_bin_to_str_1(*str++);
}

void
r_asc_str_to_bin(unsigned char * result,unsigned char * str)
{
	for ( ; *str ; )
		*result++ = r_asc_str_to_bin_1(*str++);
}




