/**********************************************************************
 
	Copyright (C) 2007 Hirohisa MORI <joshua@globalbase.org>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/

#include	<unistd.h>
#include	"memory_debug.h"
#include	"xlerror.h"
#include	"authentification.h"

AUTH_TYPE * at_list_new(AUTH_SET *,void*);
void at_list_ctl(AUTH_TYPE *,int,void*);
XL_SEXP * at_list_to_sexp(AUTH_TYPE * at);
AUTH_RESULT *
at_list_authentification(AUTH_TYPE * target,AUTH_CONNECTION_SET * set,int level,AUTH_RESULT * res);
AUTH_TYPE *
at_list_to_type(AUTH_SET * set,XL_SEXP * s);

SEM auth_lock;
AUTH_TABLE att[] = {
	{AT_AND,"and",
		at_list_new,at_list_to_sexp,at_list_to_type,0,at_list_ctl,
		0,0,0,0,0,
		at_list_authentification},
	{AT_OR,"or",
		at_list_new,at_list_to_sexp,at_list_to_type,0,at_list_ctl,
		0,0,0,0,0,
		at_list_authentification},
	{AT_TMP,0,
		0,0,0,0,0,
		0,0,0,0,0,
		0}
};


AUTH_TABLE *
at_tbl(int id)
{
AUTH_TABLE * tbl;
	for ( tbl = &att[0] ; tbl->id != AT_TMP ; tbl ++ ) {
		if ( tbl->id == id )
			return tbl;
	}
	return 0;
}

AUTH_TABLE * 
at_tbl_by_name(L_CHAR * name)
{
char * n;
AUTH_TABLE * tbl;
	n = n_string(std_cm,name);
	for ( tbl = &att[0] ; tbl->name && strcmp(tbl->name,n) ; tbl ++ );
	if ( tbl->name )
		return tbl;
	return 0;
}

AUTH_TYPE *
at_sexp_to_type(AUTH_SET * set,XL_SEXP * s)
{
AUTH_TABLE * t;
XL_SEXP * sym;
L_CHAR * n;
	sym = car(s);
	if ( get_type(sym) != XLT_SYMBOL )
		return 0;
	n = get_sf_attribute(sym->symbol.field,l_string(std_cm,"type"));
	if ( n == 0 )
		return 0;
	t = at_tbl_by_name(n);
	if ( t == 0 )
		return 0;
	return (*t->to_type)(set,s);
}

void
at_init_header(AUTH_TYPE_HEADER * h,AUTH_STATUS * sts)
{
	h->id = -1;
	h->tbl = at_tbl(h->type);
	h->sts = *sts;
}


AUTH_DIR *
_parse_xlaccess(AUTH_SET * set,AUTH_DIR * a,L_CHAR * path)
{
L_CHAR * _path;
L_CHAR * _p;
int len1,len2;
STREAM * st;
AUTH_DIR * ret;
L_CHAR * tp;
XL_SEXP *xla,*t, * sym;
L_CHAR * pf;
AUTH_PREFIX * apf,**apfp;
	tp = 0;
	_path = 0;
	_p = 0;
	if ( path == 0 || a == 0 ) {
		_path = nl_copy_str(std_cm,"/.xlaccess");
		tp = nl_copy_str(std_cm,"/");
	}
	else {
		_path = ll_copy_str(a->total_path);
		len1 = l_strlen(_path);
		len2 = l_strlen(path);
		_path = d_re_alloc(_path,sizeof(L_CHAR)*(len1+len2+50));
		l_strcpy(&_path[len1],path);
		_path[len1+len2] = '/';
		_path[len1+len2+1] = 0;
		tp = ll_copy_str(_path);
		
		l_strcpy(&_path[len1+len2+1],l_string(std_cm,".xlaccess"));
	}
	_p = get_xldocs(_path);
	st = s_open_file(n_string(std_cm,_p),O_RDONLY,0);
	if ( st == 0 ) {
		d_f_ree(_p);
		_p = get_xlsys(_path);
		st = s_open_file(n_string(std_cm,_p),O_RDONLY,0);
		if ( st == 0 ) {
			d_f_ree(_path);
			goto nothing;
		}
	}
	ret = d_alloc(sizeof(*ret));
	memset(ret,0,sizeof(*ret));
	if ( path )
		ret->dir_name = ll_copy_str(path);
	else	ret->dir_name = 0;
	ret->total_path = tp;
	tp = 0;

	xla = init_parse(st,l_string(std_cm,".xlaccess"),l_string(std_cm,".xlaccess"));
	set_encoding(xla);
	xla = get_el(xla,1);
	if ( get_type(xla) != XLT_PAIR ) {
		s_close(st);
		goto nothing;
	}
	apfp = &ret->prefix;
	for ( ; get_type(xla) == XLT_PAIR ; xla = cdr(xla) ) {
		t = car(xla);
		if ( get_type(t) != XLT_PAIR )
			continue;
		sym = car(t);
		if ( get_type(sym) != XLT_SYMBOL )
			continue;
		if ( l_strcmp(sym->symbol.data,l_string(std_cm,"at")) == 0 ) {
			pf = get_sf_attribute(sym->symbol.field,l_string(std_cm,"prefix"));
			apf = d_alloc(sizeof(*apf));
			memset(apf,0,sizeof(*apf));
			if ( pf )
				apf->prefix = ll_copy_str(pf);
			apf->at_type = at_sexp_to_type(set,t);
			if ( apf->at_type == 0 ) {
				if ( apf->prefix )
					d_f_ree(apf->prefix);
				d_f_ree(apf);
				continue;
			}
			*apfp = apf;
			apfp = &apf->next;
		}
	}
	goto end;
nothing:
	ret = d_alloc(sizeof(*ret));
	memset(ret,0,sizeof(*ret));
	if ( path )
		ret->dir_name = ll_copy_str(path);
	else	ret->dir_name = 0;
	ret->total_path = tp;
	tp = 0;
end:
	if ( tp )
		d_f_ree(tp);
	if ( _path )
		d_f_ree(_path);
	if ( _p )
		d_f_ree(_p);
	return ret;
}

XL_SEXP *
at_result_to_sexp(AUTH_RESULT * r)
{
XL_SEXP * ret,*rr;
	ret = 0;
	for ( ; r ; r = r->h.next ) {
		rr = (r->h.type->h.tbl->to_result_sexp)(r);
		ret = cons(rr,ret);
	}
	return reverse(ret);
}

AUTH_RESULT *
at_sexp_to_result(AUTH_SET * set,XL_SEXP * r)
{
AUTH_RESULT * ret, ** rp,*_ret;
XL_SEXP * t,*sym;
L_CHAR * type1;
XL_SEXP * type2;
AUTH_TABLE * tbl;
AUTH_TYPE * at;
	ret = 0;
	rp = &ret;
	for ( ; get_type(r) == XLT_PAIR ; r = cdr(r) ) {
		t = car(r);
		if ( get_type(t) != XLT_PAIR )
			continue;
		sym = car(t);
		type1 = get_sf_attribute(sym->symbol.field,l_string(std_cm,"type"));
		if ( type1 == 0 ) {
			type2 = get_el_by_symbol(t,l_string(std_cm,"at"),0);
			if ( type2 == 0 )
				continue;
			at = at_sexp_to_type(set,type2);
			_ret = (*at->h.tbl->to_result)(at,t);
		}
		else {
			tbl = at_tbl_by_name(type1);
			_ret = (tbl->to_result)(0,t);
		}
		*rp = _ret;
		_ret->h.next = 0;
		rp = &ret->h.next;
	}
	return ret;
}

AUTH_RESULT *
_get_client_result(char * pf,AUTH_SET * set)
{
int len;
char * buffer;
char * filename;
DIR * dd;
AUTH_RESULT * ret,*_ret;
struct dirent * dr;
STREAM * st;
XL_SEXP * data,*t,*sym;
	ret = 0;
	len = strlen(pf);
	buffer = d_alloc(len+100);
	filename = d_alloc(len+200);
	sprintf(buffer,"%s/authentification/clientresult",pf);
	dd = u_opendir(buffer);
	if ( dd == 0 )
		goto end1;
	for ( ; ; ) {
		dr = u_readdir(dd);
		if ( dr == 0 )
			break;
		if ( dr->d_name[0] == '.' )
			continue;
		sprintf(filename,"%s/%s",buffer,dr->d_name);
		st = s_open_file(filename,O_RDONLY);
		if ( st == 0 )
			continue;
		data = init_parse(st,l_string(std_cm,"_get_client_result"),l_string(std_cm,"_get_client_result"));
		set_encoding(data);
		data = get_el(data,1);
		for ( ; get_type(data) == XLT_PAIR ; data = cdr(data) ) {
			t = car(data);
			if ( get_type(t) != XLT_PAIR )
				continue;
			sym = car(t);
			if ( get_type(sym) != XLT_SYMBOL )
				continue;
			if ( l_strcmp(sym->symbol.data,l_string(std_cm,"result")) )
				continue;
			_ret = at_sexp_to_result(set,t);
			if ( _ret == 0 )
				continue;
			_ret->h.next = ret;
			ret = _ret;
		}
	}
	closedir(dd);
end1:
	return ret;
}

void
set_client_result(AUTH_SET * set)
{
char * pf;
AUTH_RESULT * r,**rp;
	if ( set->enable_client_result )
		return;
	pf = get_preference_path();
	if ( pf == 0 )
		goto next;
	r = _get_client_result(pf,set);
	rp = &set->client_result;
	for ( ; *rp ; rp = &(*rp)->h.next );
	*rp = r;
next:
	pf = get_peripheral_path();
	r = _get_client_result(pf,set);
	rp = &set->client_result;
	for ( ; *rp ; rp = &(*rp)->h.next );
	*rp = r;
	set->enable_client_result = 1;
}

void
free_auth_result(AUTH_RESULT * r)
{
AUTH_RESULT * r2;
	for ( ; r ; ) {
		r2 = r->h.next;
		(*r->h.type->h.tbl->free_result)(r);
	}
}


AUTH_TYPE *
at_list_new(AUTH_SET *set,void * d)
{
AUTH_TYPE * ret;
AT_LIST_T * dp;
	dp = (AT_LIST_T*)d;
	ret = d_alloc(sizeof(*ret));
	at_init_header(&ret->h,&dp->sts);
	if ( dp->d ) {
		ret->lst.len = 1;
		ret->lst.list = d_alloc(sizeof(AUTH_TYPE*));
		ret->lst.list[0] = dp->d;
	}
	else {
		ret->lst.len = 0;
		ret->lst.list = 0;
	}
	return ret;
}

XL_SEXP *
at_list_to_sexp(AUTH_TYPE * at)
{
XL_SEXP * ret,*sym;
int i;
int len,ptr;
	switch ( at->lst.len ) {
	case 0:
		return 0;
	case 1:
		return (*at->lst.list[0]->h.tbl->to_type_sexp)(at->lst.list[0]);
	default:
		len = 0;
		ptr = -1;
		for ( i = 0 ; i < at->lst.len; i ++ )
			if ( at->lst.list[i] ) {
				ptr = i;
				len ++;
			}
		switch ( len ) {
		case 0:
			return 0;
		case 1:
			return (*at->lst.list[0]->h.tbl->to_type_sexp)(at->lst.list[ptr]);
		default:
			break;
		}
		ret = 0;
		for ( i = 0 ; i < at->lst.len ; i ++ ) {
			ret = cons(
				(*at->lst.list[0]->h.tbl->to_type_sexp)(at->lst.list[i]),
				ret);
		}
		ret = reverse(ret);
		sym = n_get_symbol("auth-type");
		set_attribute(sym,l_string(std_cm,"type"),l_string(std_cm,at->h.tbl->name));
		return cons(sym,ret);
	}
}

AUTH_TYPE *
at_list_to_type(AUTH_SET * set,XL_SEXP * s)
{
AT_LIST_T att;
AUTH_TYPE *ret;
	memset(&att,0,sizeof(att));
	ret = at_list_new(set,&att);
	for ( s = cdr(s) ; get_type(s) == XLT_PAIR ; s = cdr(s) )
		(*ret->h.tbl->ctl_type)(ret,AT_CTL_INSERT,at_sexp_to_type(set,car(s)));
	return ret;
}

void
at_list_ctl(AUTH_TYPE * target,int ctl,void * d)
{
AUTH_TYPE * ins;
	switch ( ctl ) {
	case AT_CTL_INSERT:
		ins = (AUTH_TYPE*)d;
		if ( target->lst.len == 0 ) {
		}
		else {
			if ( target->lst.list[target->lst.len-1] == 0 ) {
				target->lst.list[target->lst.len-1] = ins;
			}
			else {
				target->lst.len ++;
				target->lst.list = d_re_alloc(target->lst.list,target->lst.len*sizeof(AUTH_TYPE*));
				target->lst.list[target->lst.len-1] = ins;
			}
		}
		break;
	default:
		er_panic("at_list_ctl");
	}
}

AUTH_RESULT *
at_list_authentification(AUTH_TYPE * target,AUTH_CONNECTION_SET * set,int level,AUTH_RESULT * res)
{
int i;
AUTH_RESULT * ret, *_ret,** rp;
	ret = 0;
	switch ( level ) {
	case 0:
		for ( i = 0 ; i < target->lst.len ; i ++ ) {
			if ( target->lst.list[i] == 0 )
				continue;
			_ret = (*target->lst.list[i]->h.tbl->authentification)(target->lst.list[i],set,level,res);
			if ( _ret )
				target->lst.list[i] = 0;
		}
		break;
	case 1:
		ret = 0;
		rp = &ret;
		for ( i = 0 ; i < target->lst.len ; i ++ ) {
			if ( target->lst.list[i] == 0 )
				continue;
			_ret = (*target->lst.list[i]->h.tbl->authentification)(target->lst.list[i],set,level,res);
			if ( _ret == 0 ) {
				switch ( target->h.type ) {
				case AT_AND:
					free_auth_result(ret);
					ret = 0;
					break;
				case AT_OR:
					break;
				default:
					er_panic("at_list_authentification");
				}
			}
			else {
				*rp = _ret;
				for ( ; *rp ; rp = &(*rp)->h.next);
			}
		}
		break;
	default:
		er_panic("at_list_authentification");
	}
	return ret;
}


AUTH_RESULT *
_exec_authentification_4(AUTH_TYPE * at,AUTH_RESULT * r)
{
AUTH_RESULT * ret;
	for ( ret = r ; ret ; ret = ret->h.next ) 
		if ( ret->h.type == at )
			return ret;
	return 0;
}

AUTH_RESULT *
_exec_authentification_3(AUTH_TYPE ** at_ret,AUTH_TYPE * at,AUTH_RESULT * r)
{
int i;
AUTH_RESULT * _ret,*rr;
AUTH_TYPE * _at_ret,*_at_ret2;
AT_LIST_T att;
	_ret = 0;
	switch ( at->h.type ) {
	case AT_AND:
		_at_ret = 0;
		for ( i = 0 ; i < at->lst.len ; i ++ ) {
			rr = _exec_authentification_3(&_at_ret2,at->lst.list[i],r);
			if ( rr == 0 ) {
				if ( _at_ret == 0 )
					_at_ret = _at_ret2;
				else if ( _at_ret->h.type == AT_AND ) {
					(*_at_ret->h.tbl->ctl_type)(_at_ret,AT_CTL_INSERT,_at_ret2);
				}
				else {
					memset(&att,0,sizeof(att));
					att.d = _at_ret2;
					_at_ret = (*at_tbl(AT_AND)->new_type)(0,&att);
				}
			}
			else	_ret = rr;
		}
		*at_ret = _at_ret;
		break;
	case AT_OR:
		_ret = 0;
		for ( i = 0 ; i < at->lst.len ; i ++ ) {
			_ret = _exec_authentification_3(at_ret,at->lst.list[i],r);
			if ( _ret )
				break;
		}
		if ( _ret == 0 )
			*at_ret = at;
		break;
	default:
		_ret = _exec_authentification_4(at,r);
		*at_ret = at;
	}
	return _ret;
}



AUTH_TYPE *
_exec_authentification_2(AUTH_RESULT * r,AUTH_DIR * ad,L_CHAR * prefix)
{
AUTH_PREFIX * pf;
AUTH_TYPE * ret;
	for ( pf = ad->prefix ; pf ; pf = pf->next ) {
		if ( pf->prefix == 0 ) {
			if ( _exec_authentification_3(&ret,pf->at_type,r) == 0 )
				return ret;
			break;
		}
		if ( prefix && l_strcmp(pf->prefix,prefix) )
			continue;
		if ( _exec_authentification_3(&ret,pf->at_type,r) == 0 )
			return ret;
	}
	return 0;
}

AUTH_TYPE *
_exec_authentification(AUTH_CONNECTION_SET * c_set,L_CHAR * path)
{
L_CHAR * _path;
L_CHAR ** _path_list;
int p_len;
int i,j,last_p;
AUTH_DIR * ad,*ad2;
L_CHAR * prefix;
AUTH_TYPE * ret,*rr;
AT_LIST_T att;
	_path = ll_copy_str(path);
	for ( i = l_strlen(_path)-1 ; i >= 0 ; i -- )
		if ( _path[i] == '.' || _path[i] == '/' )
			break;
	if ( i < 0 ) {
		prefix = 0;
		last_p = 0;
	}
	else if ( _path[i] == '.' ) {
		prefix = &_path[i];
		for ( last_p = i-1 ; last_p >= 0 ; last_p -- )
			if ( _path[last_p] == '/' )
				break;
		if ( last_p < 0 )
			last_p = 0;
	}
	else if ( _path[i] == '/' ) {
		prefix = 0;
		last_p = i;
	}
	p_len = 0;
	for ( i = 0 ; _path[i] ; i++ ) {
		if ( _path[i] == '/' )
			p_len ++;
	}
	p_len += 5;
	_path_list = d_alloc(sizeof(L_CHAR*)*p_len);
	j= 0;
	_path_list[j++] = _path;
	for ( i = 0 ; i <= last_p ; i ++ ) {
		if ( _path[i] == '/' ) {
			_path_list[j++] = &_path[i+1];
			_path[i] = 0;
		}
	}
	_path_list[j] = 0;
	i = 0;
	ret = 0;
	if ( c_set->base_set->auth_dir_tree == 0 )
		c_set->base_set->auth_dir_tree = _parse_xlaccess(c_set->base_set,0,0);
	for ( ad = c_set->base_set->auth_dir_tree ; ad ; ) {
		rr = _exec_authentification_2(c_set->result,ad,prefix);
		if ( rr ) {
			if ( ret == 0 )
				ret = rr;
			else if ( ret->h.type != AT_AND ) {
				att.d = rr;
				memset(&att.sts,0,sizeof(att.sts));
				ret = (*at_tbl(AT_AND)->new_type)(0,&att);
			}
			else {
				(*ret->h.tbl->ctl_type)(ret,AT_CTL_INSERT,rr);
			}
		}
		for ( ; _path_list[i] && _path_list[i][0] == 0 ; i ++ );
		if ( _path_list[i] == 0 )
			break;
		for ( ad2 = ad->child_list ; ad2  ; ad2 = ad2->child_next ) {
			if ( l_strcmp(ad2->dir_name,_path_list[i]) == 0 )
				break;
		}
		if ( ad2 == 0 ) {
			ad2 = _parse_xlaccess(c_set->base_set,ad,_path_list[i]);
			if ( ad2 ) {
				ad2->child_next = ad->child_list;
				ad->child_list = ad2;
			}
		}
		if ( ad2 == 0 )
			break;
		ad = ad2;
	}
	d_f_ree(_path_list);
	d_f_ree(_path);
	return ret;
}


XL_SEXP *
_authentification_level_0(AUTH_CONNECTION_SET * c_set,L_CHAR * path)
{
AUTH_TYPE * ret;
XL_SEXP * ret_sexp;
	ret = _exec_authentification(c_set,path);
	if ( ret == 0 )
		return 0;
	(*ret->h.tbl->authentification)(ret,c_set,0,0);
	ret_sexp = (*ret->h.tbl->to_type_sexp)(ret);
	(*ret->h.tbl->free_type)(ret,AT_FREE_TYPE_DUMMY);
	return ret_sexp;
}


XL_SEXP *
authentification_level_0(AUTH_CONNECTION_SET * c_set,L_CHAR * path)
{
XL_SEXP * ret;
	lock_task(auth_lock);
	ret = _authentification_level_0(c_set,path);
	unlock_task(auth_lock,"auth");
	return ret;
}

XL_SEXP *
_authentification_level_0_reply(AUTH_CONNECTION_SET * c_set,XL_SEXP * required)
{
AUTH_TYPE * at_req;
AUTH_RESULT * res,* res2;
XL_SEXP * ret;
	set_client_result(c_set->base_set);
	at_req = at_sexp_to_type(c_set->base_set,required);
	if ( at_req == 0 )
		return 0;
	for ( res = c_set->base_set->client_result ; res ; res = res->h.next ) {
		res2 = (*at_req->h.tbl->authentification)(at_req,c_set,1,res);
		if ( res2 )
			break;
	}
	if ( res2 == 0 ) {
		return get_error(
			required->h.file,
			required->h.line,
			XLE_PROTO_PERMISSION_DENIED,
			l_string(std_cm,"authentification level 0 reply"),
			0);
	}
	ret = at_result_to_sexp(res2);
	free_auth_result(res2);
	return ret;
}

XL_SEXP *
authentification_level_0_reply(AUTH_CONNECTION_SET * c_set,XL_SEXP * required)
{
XL_SEXP * ret;
	lock_task(auth_lock);
	ret = _authentification_level_0_reply(c_set,required);
	unlock_task(auth_lock,"level_0_reply");
	return ret;
}

XL_SEXP * 
_authentification_level_2(AUTH_CONNECTION_SET * c_set,XL_SEXP * result)
{
AUTH_RESULT * res,*rp,*res2;
XL_SEXP * ret;
	res = at_sexp_to_result(c_set->base_set,result);
	if ( res == 0 )
		return 0;
	for ( rp = res ; rp ; rp = rp->h.next ) {
		res2 = (*rp->h.type->h.tbl->authentification)(rp->h.type,c_set,2,rp);
		if ( res2 ) {
			res2->h.next = c_set->result;
			c_set->result = res2;
		}
	}
	ret = at_result_to_sexp(res);
	free_auth_result(res);
	return ret;
}

XL_SEXP *
authentification_level_2(AUTH_CONNECTION_SET * c_set,XL_SEXP * result)
{
XL_SEXP * ret;
	lock_task(auth_lock);
	ret = _authentification_level_2(c_set,result);
	unlock_task(auth_lock,"level_2");
	return ret;
}




